@@ -24,13 +24,22 @@
 
 G_BEGIN_DECLS
 
+typedef enum {
+  GCLUE_TOWER_TEC_UNKNOWN = 0,
+  GCLUE_TOWER_TEC_3G = 1,
+  GCLUE_TOWER_TEC_4G = 2,
+} GClueTowerTec;
+
 typedef struct _GClue3GTower GClue3GTower;
 
+#define GCLUE_3G_TOWER_OPERATOR_CODE_STR_LEN 6
+#define GCLUE_3G_TOWER_COUNTRY_CODE_STR_LEN 3
+
 struct _GClue3GTower {
-        guint   mcc;
-        guint   mnc;
+        gchar   opc[GCLUE_3G_TOWER_OPERATOR_CODE_STR_LEN + 1];
         gulong  lac;
         gulong  cell_id;
+        GClueTowerTec tec;
 };
 
 G_END_DECLS
