@@ -24,12 +24,14 @@
  */
 
 #include "gclue-location.h"
+#include "gclue-nmea-utils.h"
 #include <math.h>
 #include <string.h>
 #include <stdlib.h>
 
 #define TIME_DIFF_THRESHOLD 60000000 /* 60 seconds */
 #define EARTH_RADIUS_KM 6372.795
+#define KNOTS_IN_METERS_PER_SECOND 0.51444
 
 struct _GClueLocationPrivate {
         char   *description;
@@ -228,13 +230,13 @@ static void
 gclue_location_constructed (GObject *object)
 {
         GClueLocation *location = GCLUE_LOCATION (object);
-        GTimeVal tv;
+        gint64 timestamp;
 
         if (location->priv->timestamp != 0)
                 return;
 
-        g_get_current_time (&tv);
-        gclue_location_set_timestamp (location, tv.tv_sec);
+        timestamp = g_get_real_time () / G_USEC_PER_SEC;
+        gclue_location_set_timestamp (location, timestamp);
 }
 
 static void
@@ -390,9 +392,7 @@ gclue_location_class_init (GClueLocationClass *klass)
 static void
 gclue_location_init (GClueLocation *location)
 {
-        location->priv = G_TYPE_INSTANCE_GET_PRIVATE ((location),
-                                                      GCLUE_TYPE_LOCATION,
-                                                      GClueLocationPrivate);
+        location->priv = gclue_location_get_instance_private (location);
 
         location->priv->altitude = GCLUE_LOCATION_ALTITUDE_UNKNOWN;
         location->priv->accuracy = GCLUE_LOCATION_ACCURACY_UNKNOWN;
@@ -491,6 +491,9 @@ parse_nmea_timestamp (const char *nmea_ts)
         now = g_date_time_new_now_utc ();
         ret = g_date_time_to_unix (now);
 
+	if( now == NULL)
+                goto parse_error;
+
         if (strlen (nmea_ts) < 6) {
                 if (strlen (nmea_ts) >= 1)
                         /* Empty string just means no ts, so no warning */
@@ -514,6 +517,8 @@ parse_nmea_timestamp (const char *nmea_ts)
                                   hours,
                                   minutes,
                                   seconds);
+        if (ts == NULL)
+                goto parse_error;
 
         if (g_date_time_difference (ts, now) > TIME_DIFF_THRESHOLD) {
                 g_debug ("NMEA timestamp '%s' in future. Assuming yesterday's.",
@@ -595,17 +600,7 @@ gclue_location_new_full (gdouble     latitude,
                              NULL);
 }
 
-/**
- * gclue_location_create_from_gga:
- * @gga: NMEA GGA sentence
- * @error: Place-holder for errors.
- *
- * Creates a new #GClueLocation object from a GGA sentence.
- *
- * Returns: a new #GClueLocation object, or %NULL on error. Unref using
- * #g_object_unref() when done with it.
- **/
-GClueLocation *
+static GClueLocation *
 gclue_location_create_from_gga (const char *gga, GError **error)
 {
         GClueLocation *location = NULL;
@@ -656,6 +651,121 @@ out:
         return location;
 }
 
+static GClueLocation *
+gclue_location_create_from_rmc (const char     *rmc,
+                                GClueLocation  *prev_location,
+                                GError        **error)
+{
+        GClueLocation *location = NULL;
+        char **parts = g_strsplit (rmc, ",", -1);
+
+        if (g_strv_length (parts) < 13)
+                goto error;
+
+        guint64 timestamp = parse_nmea_timestamp (parts[1]);
+        gdouble lat = parse_coordinate_string (parts[3], parts[4]);
+        gdouble lon = parse_coordinate_string (parts[5], parts[6]);
+
+        if (lat == INVALID_COORDINATE || lon == INVALID_COORDINATE)
+                goto error;
+
+        gdouble speed = GCLUE_LOCATION_SPEED_UNKNOWN;
+        if (parts[7][0] != '\0')
+                speed = g_ascii_strtod (parts[7], NULL) * KNOTS_IN_METERS_PER_SECOND;
+
+        gdouble heading = GCLUE_LOCATION_HEADING_UNKNOWN;
+        if (parts[8][0] != '\0')
+                heading = g_ascii_strtod (parts[8], NULL);
+
+        /* Some receivers use '0.0,0.0' as invalid speed and heading */
+        if (speed == 0.0 && heading == 0.0) {
+                speed = GCLUE_LOCATION_SPEED_UNKNOWN;
+                heading = GCLUE_LOCATION_HEADING_UNKNOWN;
+        }
+
+        location = g_object_new (GCLUE_TYPE_LOCATION,
+                                 "latitude", lat,
+                                 "longitude", lon,
+                                 "timestamp", timestamp,
+                                 "speed", speed,
+                                 "heading", heading,
+                                 NULL);
+
+        if (prev_location != NULL) {
+                g_object_set (location,
+                              "accuracy",
+                              gclue_location_get_accuracy (prev_location),
+                              NULL);
+                g_object_set (location,
+                              "altitude",
+                              gclue_location_get_altitude (prev_location),
+                              NULL);
+        }
+
+        goto out;
+error:
+        g_set_error_literal (error,
+                             G_IO_ERROR,
+                             G_IO_ERROR_INVALID_ARGUMENT,
+                             "Invalid NMEA RMC sentence");
+out:
+        g_strfreev (parts);
+        return location;
+}
+
+/**
+ * gclue_location_create_from_nmeas:
+ * @nmea: A NULL terminated array NMEA sentence strings
+ * @prev_location: Previous location provided from the location source
+ * @error: Place-holder for errors.
+ *
+ * Creates a new #GClueLocation object by combining data from multiple NMEA
+ * sentences.
+ *
+ * Returns: a new #GClueLocation object if GGA or RMC sentences are found,
+ * a %NULL on all other cases and errors. Unref using #g_object_unref() when
+ * done with it.
+ **/
+GClueLocation *
+gclue_location_create_from_nmeas (const char     *nmeas[],
+                                  GClueLocation  *prev_location,
+                                  GError        **error)
+{
+        GClueLocation *gga_loc = NULL;
+        GClueLocation *rmc_loc = NULL;
+        const char **iter;
+
+        for (iter = nmeas; *iter != NULL; iter++) {
+                if (!gga_loc && gclue_nmea_type_is (*iter, "GGA"))
+                        gga_loc = gclue_location_create_from_gga (*iter, NULL);
+                if (!rmc_loc && gclue_nmea_type_is (*iter, "RMC"))
+                        rmc_loc = gclue_location_create_from_rmc
+                                (*iter, prev_location, NULL);
+                if (gga_loc && rmc_loc)
+                    break;
+        }
+
+        if (gga_loc && rmc_loc) {
+                gclue_location_set_speed
+                        (gga_loc, gclue_location_get_speed(rmc_loc));
+                gclue_location_set_heading
+                        (gga_loc, gclue_location_get_heading(rmc_loc));
+                g_object_unref (rmc_loc);
+
+                return gga_loc;
+        }
+        if (gga_loc)
+                return gga_loc;
+        if (rmc_loc)
+                return rmc_loc;
+
+        g_set_error_literal (error,
+                             G_IO_ERROR,
+                             G_IO_ERROR_INVALID_ARGUMENT,
+                             "Valid NMEA GGA or RMC sentence not found");
+        return NULL;
+}
+
 /**
  * gclue_location_duplicate:
  * @location: the #GClueLocation instance to duplicate.
@@ -839,8 +949,8 @@ gclue_location_set_speed_from_prev_location (GClueLocation *location,
                goto out;
         }
 
-        speed = gclue_location_get_distance_from (location, prev_location) *
-                1000.0 / (timestamp - prev_timestamp);
+        speed = gclue_location_get_distance_from (location, prev_location) /
+                (timestamp - prev_timestamp);
 
 out:
         location->priv->speed = speed;
@@ -895,7 +1005,7 @@ void
 gclue_location_set_heading_from_prev_location (GClueLocation *location,
                                                GClueLocation *prev_location)
 {
-        gdouble dx, dy, angle, lat, lon, prev_lat, prev_lon;
+        gdouble dlat, dlon, x, y, angle, lat, lon, prev_lat, prev_lon;
 
         g_return_if_fail (GCLUE_IS_LOCATION (location));
         g_return_if_fail (prev_location == NULL ||
@@ -912,35 +1022,42 @@ gclue_location_set_heading_from_prev_location (GClueLocation *location,
         prev_lat = gclue_location_get_latitude (prev_location);
         prev_lon = gclue_location_get_longitude (prev_location);
 
-        dx = (lat - prev_lat);
-        dy = (lon - prev_lon);
+        if (lat == prev_lat && lon == prev_lon) {
+               location->priv->heading = GCLUE_LOCATION_HEADING_UNKNOWN;
 
-        /* atan2 takes in coordinate values of a 2D space and returns the angle
-         * which the line from origin to that coordinate makes with the positive
-         * X-axis, in the range (-PI,+PI]. Converting it into degrees we get the
-         * angle in range (-180,180]. This means East = 0 degree,
-         * West = -180 degrees, North = 90 degrees, South = -90 degrees.
+               return;
+        }
+
+        /* atan2(y, x) is a function which takes in coordinate values of
+         * a 2D point and returns the angle of line from origin to that
+         * coordinate makes with the positive X-axis, in the range (-PI,+PI].
          *
-         * Passing atan2 a negative value of dx will flip the angles about
-         * Y-axis. This means the angle now returned will be the angle with
-         * respect to negative X-axis. Which makes West = 0 degree,
-         * East = 180 degrees, North = 90 degrees, South = -90 degrees. */
-        angle = atan2(dy, -dx) * 180.0 / M_PI;
-
-        /* Now, North is supposed to be 0 degree. Lets subtract 90 degrees
-         * from angle. After this step West = -90 degrees, East = 90 degrees,
-         * North = 0 degree, South = -180 degrees. */
-        angle -= 90.0;
-
-        /* As we know, angle ~= angle + 360; using this on negative values would
-         * bring the the angle in range [0,360).
+         * We want to match the inclusive end (+PI) of atan2's range to our
+         * desired zero direction (north), so we take the vector pointing
+         * south as our X-axis and the vector pointing east as our Y-axis
+         * and measure our movement vector coordinates */
+        dlon = (lon - prev_lon);
+        dlat = (lat - prev_lat);
+        /* in that basis.
          *
-         * After this step West = 270 degrees, East = 90 degrees,
-         * North = 0 degree, South = 180 degrees. */
-        if (angle < 0)
-                angle += 360.0;
-
-        location->priv->heading = angle;
+         * Latitudes decrease towards the south and longitudes
+         * increase towards the east, so: */
+        x = -dlat;
+        y = dlon;
+        angle = atan2(y, x) * 180.0 / M_PI;
+        /* We now have the angle of our movement vector measured from the
+         * vector pointing south in degrees, with the positive direction
+         * being counterclockwise.
+         *
+         * We want to convert it to heading: Degrees from from north with
+         * positive values only and positive direction being clockwise.
+         *
+         * If angle is positive (the movement is more towards east than
+         * west), we need to subtract it from the heading of our reference
+         * vector (south == 180 deg). If the angle is negative, we need to
+         * add its negative to the heading of the reference vector. Both
+         * cases result in */
+        location->priv->heading = 180.0 - angle;
 
         g_object_notify (G_OBJECT (location), "heading");
 }
@@ -950,11 +1067,11 @@ gclue_location_set_heading_from_prev_location (GClueLocation *location,
  * @loca: a #GClueLocation
  * @locb: a #GClueLocation
  *
- * Calculates the distance in km, along the curvature of the Earth,
+ * Calculates the distance in meters, along the curvature of the Earth,
  * between 2 locations. Note that altitude changes are not
  * taken into account.
  *
- * Returns: a distance in km.
+ * Returns: a distance in meters.
  **/
 double
 gclue_location_get_distance_from (GClueLocation *loca,
@@ -977,5 +1094,5 @@ gclue_location_get_distance_from (GClueLocation *loca,
         a = sin (dlat / 2) * sin (dlat / 2) +
             sin (dlon / 2) * sin (dlon / 2) * cos (lat1) * cos (lat2);
         c = 2 * atan2 (sqrt (a), sqrt (1-a));
-        return EARTH_RADIUS_KM * c;
+        return 1000.0 * EARTH_RADIUS_KM * c;
 }
