@@ -31,11 +31,6 @@
 #include "gclue-locator.h"
 #include "gclue-config.h"
 
-/* 20 seconds as milliseconds */
-#define AGENT_WAIT_TIMEOUT 20000
-/*  20 seconds as microseconds */
-#define AGENT_WAIT_TIMEOUT_USEC (20 * G_USEC_PER_SEC)
-
 static void
 gclue_service_manager_manager_iface_init (GClueDBusManagerIface *iface);
 static void
@@ -46,10 +41,10 @@ struct _GClueServiceManagerPrivate
         GDBusConnection *connection;
         GList *clients;
         GHashTable *agents;
+        GQueue *clients_waiting_agent;
 
         guint last_client_id;
         guint num_clients;
-        gint64 init_time;
 
         GClueLocator *locator;
 };
@@ -73,6 +68,21 @@ enum
 
 static GParamSpec *gParamSpecs[LAST_PROP];
 
+typedef struct
+{
+        GClueDBusManager *manager;
+        GDBusMethodInvocation *invocation;
+        GClueClientInfo *client_info;
+        gboolean reuse_client;
+
+} OnClientInfoNewReadyData;
+
+static void
+on_client_info_new_ready_data_free (gpointer data)
+{
+        g_slice_free (OnClientInfoNewReadyData, data);
+}
+
 static void
 sync_in_use_property (GClueServiceManager *manager)
 {
@@ -127,7 +137,7 @@ delete_client (GClueServiceManager *manager,
 
                 if (compare_func (l->data, compare_func_data) == 0) {
                         g_object_unref (G_OBJECT (l->data));
-                        priv->clients = g_list_remove_link (priv->clients, l);
+                        priv->clients = g_list_delete_link (priv->clients, l);
                         priv->num_clients--;
                         if (priv->num_clients == 0) {
                                 g_object_notify (G_OBJECT (manager), "active");
@@ -140,6 +150,14 @@ delete_client (GClueServiceManager *manager,
         sync_in_use_property (manager);
 }
 
+static void
+on_client_notify_active (GObject    *gobject,
+                         GParamSpec *pspec,
+                         gpointer    user_data)
+{
+        sync_in_use_property (GCLUE_SERVICE_MANAGER (user_data));
+}
+
 static void
 on_peer_vanished (GClueClientInfo *info,
                   gpointer         user_data)
@@ -155,15 +173,6 @@ on_peer_vanished (GClueClientInfo *info,
                        (char *) bus_name);
 }
 
-typedef struct
-{
-        GClueDBusManager *manager;
-        GDBusMethodInvocation *invocation;
-        GClueClientInfo *client_info;
-        gboolean reuse_client;
-
-} OnClientInfoNewReadyData;
-
 static gboolean
 complete_get_client (OnClientInfoNewReadyData *data)
 {
@@ -175,6 +184,9 @@ complete_get_client (OnClientInfoNewReadyData *data)
         char *path;
         guint32 user_id;
 
+        /* Disconnect on_peer_vanished_before_completion, if it's there */
+        g_signal_handlers_disconnect_by_data (info, data);
+
         user_id = gclue_client_info_get_user_id (info);
         agent_proxy = g_hash_table_lookup (priv->agents,
                                            GINT_TO_POINTER (user_id));
@@ -216,6 +228,11 @@ complete_get_client (OnClientInfoNewReadyData *data)
         }
         g_debug ("Number of connected clients: %u", priv->num_clients);
 
+        g_signal_connect (client,
+                          "notify::active",
+                          G_CALLBACK (on_client_notify_active),
+                          data->manager);
+
         g_signal_connect (info,
                           "peer-vanished",
                           G_CALLBACK (on_peer_vanished),
@@ -240,12 +257,36 @@ error_out:
 out:
         g_clear_error (&error);
         g_clear_object (&info);
-        g_slice_free (OnClientInfoNewReadyData, data);
+        on_client_info_new_ready_data_free (data);
         g_free (path);
 
         return FALSE;
 }
 
+static void
+on_peer_vanished_before_completion (GClueClientInfo *info,
+                                    gpointer         user_data)
+{
+        OnClientInfoNewReadyData *data = (OnClientInfoNewReadyData *) user_data;
+        GClueServiceManager *self = GCLUE_SERVICE_MANAGER (data->manager);
+        GClueServiceManagerPrivate *priv = self->priv;
+        const char *bus_name;
+
+        bus_name = gclue_client_info_get_bus_name (info);
+        g_debug ("Client `%s` vanished before agent appeared",
+                 bus_name);
+        g_signal_handlers_disconnect_by_data (info,
+                                              user_data);
+        g_dbus_method_invocation_return_error (data->invocation,
+                                               G_DBUS_ERROR,
+                                               G_DBUS_ERROR_DISCONNECTED,
+                                               "%s vanished before completion",
+                                               bus_name);
+        g_queue_remove (priv->clients_waiting_agent, data);
+        on_client_info_new_ready_data_free (data);
+        g_object_unref (info);
+}
+
 static void
 on_client_info_new_ready (GObject      *source_object,
                           GAsyncResult *res,
@@ -257,8 +298,6 @@ on_client_info_new_ready (GObject      *source_object,
         GClueAgent *agent_proxy;
         GError *error = NULL;
         guint32 user_id;
-        gint64 now;
-        gboolean system_app;
 
         info = gclue_client_info_new_finish (res, &error);
         if (info == NULL) {
@@ -267,7 +306,7 @@ on_client_info_new_ready (GObject      *source_object,
                                                                G_DBUS_ERROR_FAILED,
                                                                error->message);
                 g_error_free (error);
-                g_slice_free (OnClientInfoNewReadyData, data);
+                on_client_info_new_ready_data_free (data);
 
                 return;
         }
@@ -277,19 +316,20 @@ on_client_info_new_ready (GObject      *source_object,
         user_id = gclue_client_info_get_user_id (info);
         agent_proxy = g_hash_table_lookup (priv->agents,
                                            GINT_TO_POINTER (user_id));
-        now = g_get_monotonic_time ();
-
-        system_app = (gclue_client_info_get_xdg_id (info) == NULL);
-        if (agent_proxy == NULL &&
-            !system_app &&
-            now < (priv->init_time + AGENT_WAIT_TIMEOUT_USEC)) {
+        if (agent_proxy == NULL) {
                 /* Its possible that geoclue was just launched on GetClient
                  * call, in which case agents need some time to register
                  * themselves to us.
                  */
-                g_timeout_add (AGENT_WAIT_TIMEOUT,
-                               (GSourceFunc) complete_get_client,
-                               user_data);
+                g_queue_push_tail (priv->clients_waiting_agent, data);
+                g_signal_connect
+                        (info,
+                         "peer-vanished",
+                         G_CALLBACK (on_peer_vanished_before_completion),
+                         user_data);
+                g_debug ("Client `%s` waiting for agent for user ID '%u'",
+                         gclue_client_info_get_bus_name (info), user_id);
+
                 return;
         }
 
@@ -420,6 +460,7 @@ on_agent_proxy_ready (GObject      *source_object,
         guint32 user_id;
         GClueAgent *agent;
         GError *error = NULL;
+        GList *l;
 
         agent = gclue_agent_proxy_new_for_bus_finish (res, &error);
         if (agent == NULL)
@@ -436,6 +477,19 @@ on_agent_proxy_ready (GObject      *source_object,
 
         gclue_dbus_manager_complete_add_agent (data->manager, data->invocation);
 
+        l = priv->clients_waiting_agent->head;
+        while (l != NULL) {
+                GList *next = l->next;
+                OnClientInfoNewReadyData *d =
+                        (OnClientInfoNewReadyData *) l->data;
+
+                if (gclue_client_info_get_user_id(d->client_info) == user_id) {
+                        complete_get_client (d);
+                        g_queue_delete_link (priv->clients_waiting_agent, l);
+                }
+                l = next;
+        }
+
         goto out;
 
 error_out:
@@ -533,6 +587,11 @@ gclue_service_manager_finalize (GObject *object)
                 priv->clients = NULL;
         }
         g_clear_pointer (&priv->agents, g_hash_table_unref);
+        if (priv->clients_waiting_agent != NULL) {
+                g_queue_free_full (priv->clients_waiting_agent,
+                                   on_client_info_new_ready_data_free);
+                priv->clients_waiting_agent = NULL;
+        }
 
         /* Chain up to the parent class */
         G_OBJECT_CLASS (gclue_service_manager_parent_class)->finalize (object);
@@ -649,15 +708,13 @@ gclue_service_manager_class_init (GClueServiceManagerClass *klass)
 static void
 gclue_service_manager_init (GClueServiceManager *manager)
 {
-        manager->priv = G_TYPE_INSTANCE_GET_PRIVATE (manager,
-                                                     GCLUE_TYPE_SERVICE_MANAGER,
-                                                     GClueServiceManagerPrivate);
+        manager->priv = gclue_service_manager_get_instance_private (manager);
 
         manager->priv->agents = g_hash_table_new_full (g_direct_hash,
                                                        g_direct_equal,
                                                        NULL,
                                                        g_object_unref);
-        manager->priv->init_time = g_get_monotonic_time ();
+        manager->priv->clients_waiting_agent = g_queue_new ();
 }
 
 static gboolean
