@@ -34,10 +34,21 @@
  */
 #define WIFI_SCAN_TIMEOUT_LOW_ACCURACY  300
 
-#define BSSID_LEN 7
-#define BSSID_STR_LEN 18
+/* WiFi APs at and below this signal level in scan results are ignored.
+ * In dBm units.
+ */
+#define WIFI_SCAN_BSS_NOISE_LEVEL -90
+
+#define BSSID_LEN 6
+#define BSSID_STR_LEN 17
 #define MAX_SSID_LEN 32
 
+/* Drop entries from the cache when they are more than 48 hours old. If we are
+ * polling at high accuracy for that entire period, that gives a maximum cache
+ * size of 17280 entries. At roughly 400B each, that’s about 7MB of heap for a
+ * full cache (excluding overheads). */
+#define CACHE_ENTRY_MAX_AGE_SECONDS (48 * 60 * 60)
+
 /**
  * SECTION:gclue-wifi
  * @short_description: WiFi-based geolocation
@@ -46,11 +57,27 @@
  * Contains functions to get the geolocation based on nearby WiFi networks.
  **/
 
-static gboolean
+static GClueLocationSourceStartResult
 gclue_wifi_start (GClueLocationSource *source);
-static gboolean
+static GClueLocationSourceStopResult
 gclue_wifi_stop (GClueLocationSource *source);
 
+static guint
+variant_hash (gconstpointer key);
+
+static void
+gclue_wifi_refresh_async (GClueWebSource      *source,
+                          GCancellable        *cancellable,
+                          GAsyncReadyCallback  callback,
+                          gpointer             user_data);
+static GClueLocation *
+gclue_wifi_refresh_finish (GClueWebSource  *source,
+                           GAsyncResult    *result,
+                           GError         **error);
+
+static void
+disconnect_cache_prune_timeout (GClueWifi *wifi);
+
 struct _GClueWifiPrivate {
         WPASupplicant *supplicant;
         WPAInterface *interface;
@@ -61,10 +88,20 @@ struct _GClueWifiPrivate {
         gulong bss_added_id;
         gulong bss_removed_id;
         gulong scan_done_id;
+        guint scan_wait_id;
 
         guint scan_timeout;
 
         GClueAccuracyLevel accuracy_level;
+
+        GHashTable *location_cache;  /* (element-type GVariant GClueLocation) (owned) */
+        guint cache_prune_timeout_id;
+        guint cache_hits, cache_misses;
+
+#if GLIB_CHECK_VERSION(2, 64, 0)
+        GMemoryMonitor *memory_monitor;
+        gulong low_memory_warning_id;
+#endif
 };
 
 enum
@@ -114,10 +151,13 @@ gclue_wifi_finalize (GObject *gwifi)
         G_OBJECT_CLASS (gclue_wifi_parent_class)->finalize (gwifi);
 
         disconnect_bss_signals (wifi);
+        disconnect_cache_prune_timeout (wifi);
+
         g_clear_object (&wifi->priv->supplicant);
         g_clear_object (&wifi->priv->interface);
         g_clear_pointer (&wifi->priv->bss_proxies, g_hash_table_unref);
         g_clear_pointer (&wifi->priv->ignored_bss_proxies, g_hash_table_unref);
+        g_clear_pointer (&wifi->priv->location_cache, g_hash_table_unref);
 }
 
 static void
@@ -168,6 +208,8 @@ gclue_wifi_class_init (GClueWifiClass *klass)
 
         source_class->start = gclue_wifi_start;
         source_class->stop = gclue_wifi_stop;
+        web_class->refresh_async = gclue_wifi_refresh_async;
+        web_class->refresh_finish = gclue_wifi_refresh_finish;
         web_class->create_submit_query = gclue_wifi_create_submit_query;
         web_class->create_query = gclue_wifi_create_query;
         web_class->parse_response = gclue_wifi_parse_response;
@@ -205,7 +247,7 @@ variant_to_string (GVariant *variant, guint max_len, char *ret)
         len = g_variant_n_children (variant);
         if (len == 0)
                 return 0;
-        g_return_val_if_fail(len < max_len, 0);
+        g_return_val_if_fail(len <= max_len, 0);
         ret[len] = '\0';
 
         for (i = 0; i < len; i++)
@@ -229,7 +271,7 @@ static gboolean
 get_bssid_from_bss (WPABSS *bss, char *bssid)
 {
         GVariant *variant;
-        char raw_bssid[BSSID_LEN] = { 0 };
+        char raw_bssid[BSSID_LEN + 1] = { 0 };
         guint raw_len, i;
 
         variant = wpa_bss_get_bssid (bss);
@@ -237,12 +279,12 @@ get_bssid_from_bss (WPABSS *bss, char *bssid)
                 return FALSE;
 
         raw_len = variant_to_string (variant, BSSID_LEN, raw_bssid);
-        g_return_val_if_fail (raw_len == BSSID_LEN - 1, FALSE);
+        g_return_val_if_fail (raw_len == BSSID_LEN, FALSE);
 
-        for (i = 0; i < BSSID_LEN - 1; i++) {
+        for (i = 0; i < BSSID_LEN; i++) {
                 unsigned char c = (unsigned char) raw_bssid[i];
 
-                if (i == BSSID_LEN - 2) {
+                if (i == BSSID_LEN - 1) {
                         g_snprintf (bssid + (i * 3), 3, "%02x", c);
                 } else {
                         g_snprintf (bssid + (i * 3), 4, "%02x:", c);
@@ -262,7 +304,7 @@ add_bss_proxy (GClueWifi *wifi,
         if (g_hash_table_replace (wifi->priv->bss_proxies,
                                   g_strdup (path),
                                   bss)) {
-                char ssid[MAX_SSID_LEN] = { 0 };
+                char ssid[MAX_SSID_LEN + 1] = { 0 };
 
                 wifi->priv->bss_list_changed = TRUE;
                 get_ssid_from_bss (bss, ssid);
@@ -279,11 +321,11 @@ on_bss_signal_notify (GObject    *gobject,
         WPABSS *bss = WPA_BSS (gobject);
         const char *path;
 
-        if (wpa_bss_get_signal (bss) <= -90) {
-                char bssid[BSSID_STR_LEN] = { 0 };
+        if (wpa_bss_get_signal (bss) <= WIFI_SCAN_BSS_NOISE_LEVEL) {
+                char bssid[BSSID_STR_LEN + 1] = { 0 };
 
                 get_bssid_from_bss (bss, bssid);
-                g_debug ("WiFi AP '%s' still has very low strength (%u dBm)"
+                g_debug ("WiFi AP '%s' still has very low strength (%d dBm)"
                          ", ignoring again…",
                          bssid,
                          wpa_bss_get_signal (bss));
@@ -306,7 +348,7 @@ on_bss_proxy_ready (GObject      *source_object,
         GClueWifi *wifi = GCLUE_WIFI (user_data);
         WPABSS *bss;
         GError *error = NULL;
-        char ssid[MAX_SSID_LEN] = { 0 };
+        char ssid[MAX_SSID_LEN + 1] = { 0 };
 
         bss = wpa_bss_proxy_new_for_bus_finish (res, &error);
         if (bss == NULL) {
@@ -325,12 +367,12 @@ on_bss_proxy_ready (GObject      *source_object,
         get_ssid_from_bss (bss, ssid);
         g_debug ("WiFi AP '%s' added.", ssid);
 
-        if (wpa_bss_get_signal (bss) <= -90) {
+        if (wpa_bss_get_signal (bss) <= WIFI_SCAN_BSS_NOISE_LEVEL) {
                 const char *path;
-                char bssid[BSSID_STR_LEN] = { 0 };
+                char bssid[BSSID_STR_LEN + 1] = { 0 };
 
                 get_bssid_from_bss (bss, bssid);
-                g_debug ("WiFi AP '%s' has very low strength (%u dBm)"
+                g_debug ("WiFi AP '%s' has very low strength (%d dBm)"
                          ", ignoring for now…",
                          bssid,
                          wpa_bss_get_signal (bss));
@@ -366,7 +408,7 @@ on_bss_added (WPAInterface *object,
 static gboolean
 remove_bss_from_hashtable (const gchar *path, GHashTable *hash_table)
 {
-        char ssid[MAX_SSID_LEN] = { 0 };
+        char ssid[MAX_SSID_LEN + 1] = { 0 };
         WPABSS *bss = NULL;
 
         bss = g_hash_table_lookup (hash_table, path);
@@ -432,6 +474,11 @@ cancel_wifi_scan (GClueWifi *wifi)
                 priv->scan_timeout = 0;
         }
 
+        if (priv->scan_wait_id != 0) {
+                g_source_remove (priv->scan_wait_id);
+                priv->scan_wait_id = 0;
+        }
+
         if (priv->scan_done_id != 0) {
                 g_signal_handler_disconnect (priv->interface,
                                              priv->scan_done_id);
@@ -453,7 +500,25 @@ on_scan_timeout (gpointer user_data)
 
         start_wifi_scan (wifi);
 
-        return FALSE;
+        return G_SOURCE_REMOVE;
+}
+
+static gboolean
+on_scan_wait_done (gpointer wifi)
+{
+        GClueWifiPrivate *priv;
+
+        g_return_val_if_fail (GCLUE_IS_WIFI (wifi), G_SOURCE_REMOVE);
+        priv = GCLUE_WIFI(wifi)->priv;
+
+        if (priv->bss_list_changed) {
+                priv->bss_list_changed = FALSE;
+                g_debug ("Refreshing location…");
+                gclue_web_source_refresh (GCLUE_WEB_SOURCE (wifi));
+        }
+        priv->scan_wait_id = 0;
+
+        return G_SOURCE_REMOVE;
 }
 
 static void
@@ -475,11 +540,10 @@ on_scan_done (WPAInterface *object,
         if (priv->interface == NULL)
                 return;
 
-        if (priv->bss_list_changed) {
-                priv->bss_list_changed = FALSE;
-                g_debug ("Refreshing location…");
-                gclue_web_source_refresh (GCLUE_WEB_SOURCE (wifi));
-        }
+        if (priv->scan_wait_id != 0)
+            g_source_remove (priv->scan_wait_id);
+
+        priv->scan_wait_id = g_timeout_add_seconds (1, on_scan_wait_done, wifi);
 
         /* If there was another scan already scheduled, cancel that and
          * re-schedule. Regardless of our internal book-keeping, this can happen
@@ -586,34 +650,152 @@ disconnect_bss_signals (GClueWifi *wifi)
         g_hash_table_remove_all (priv->ignored_bss_proxies);
 }
 
+static void
+cache_prune (GClueWifi *wifi)
+{
+        GClueWifiPrivate *priv = wifi->priv;
+        GHashTableIter iter;
+        gpointer value;
+        guint64 cutoff_seconds;
+        guint old_cache_size;
+
+        old_cache_size = g_hash_table_size (priv->location_cache);
+        cutoff_seconds = g_get_real_time () / G_USEC_PER_SEC - CACHE_ENTRY_MAX_AGE_SECONDS;
+
+        g_hash_table_iter_init (&iter, priv->location_cache);
+        while (g_hash_table_iter_next (&iter, NULL, &value)) {
+                GClueLocation *location = GCLUE_LOCATION (value);
+                guint64 timestamp_seconds = gclue_location_get_timestamp (location);
+
+                if (timestamp_seconds <= cutoff_seconds)
+                        g_hash_table_iter_remove (&iter);
+        }
+
+        g_debug ("Pruned cache (old size: %u, new size: %u)",
+                 old_cache_size, g_hash_table_size (priv->location_cache));
+}
+
+#if GLIB_CHECK_VERSION(2, 64, 0)
+static void
+cache_empty (GClueWifi *wifi)
+{
+        GClueWifiPrivate *priv = wifi->priv;
+
+        g_debug ("Emptying cache");
+        g_hash_table_remove_all (priv->location_cache);
+}
+#endif  /* GLib ≥ 2.64.0 */
+
 static gboolean
+cache_prune_timeout_cb (gpointer user_data)
+{
+        GClueWifi *wifi = GCLUE_WIFI (user_data);
+
+        cache_prune (wifi);
+
+        return G_SOURCE_CONTINUE;
+}
+
+#if GLIB_CHECK_VERSION(2, 64, 0)
+static void
+low_memory_warning_cb (GMemoryMonitor             *memory_monitor,
+                       GMemoryMonitorWarningLevel  level,
+                       gpointer                    user_data)
+{
+        GClueWifi *wifi = GCLUE_WIFI (user_data);
+
+        if (level == G_MEMORY_MONITOR_WARNING_LEVEL_LOW)
+                cache_prune (wifi);
+        else if (level > G_MEMORY_MONITOR_WARNING_LEVEL_LOW)
+                cache_empty (wifi);
+}
+#endif  /* GLib ≥ 2.64.0 */
+
+static void
+connect_cache_prune_timeout (GClueWifi *wifi)
+{
+        GClueWifiPrivate *priv = wifi->priv;
+
+        g_debug ("Connecting cache prune timeout");
+
+        /* Run the prune at twice the expiry frequency, which should allow us to
+         * sample often enough to keep the expiries at that frequency, as per
+         * Nyquist. */
+        if (priv->cache_prune_timeout_id != 0)
+                g_source_remove (priv->cache_prune_timeout_id);
+        priv->cache_prune_timeout_id = g_timeout_add_seconds (CACHE_ENTRY_MAX_AGE_SECONDS / 2,
+                                                              cache_prune_timeout_cb,
+                                                              wifi);
+
+#if GLIB_CHECK_VERSION(2, 64, 0)
+        if (priv->memory_monitor == NULL) {
+                priv->memory_monitor = g_memory_monitor_dup_default ();
+                priv->low_memory_warning_id = g_signal_connect (priv->memory_monitor,
+                                                                "low-memory-warning",
+                                                                G_CALLBACK (low_memory_warning_cb),
+                                                                wifi);
+        }
+#endif
+}
+
+static void
+disconnect_cache_prune_timeout (GClueWifi *wifi)
+{
+        GClueWifiPrivate *priv = wifi->priv;
+
+        g_debug ("Disconnecting cache prune timeout");
+
+        /* Run one last prune. */
+        cache_prune (wifi);
+
+#if GLIB_CHECK_VERSION(2, 64, 0)
+        if (priv->low_memory_warning_id != 0 && priv->memory_monitor != NULL)
+                g_signal_handler_disconnect (priv->memory_monitor, priv->low_memory_warning_id);
+        g_clear_object (&priv->memory_monitor);
+#endif
+
+        if (priv->cache_prune_timeout_id != 0)
+                g_source_remove (priv->cache_prune_timeout_id);
+        priv->cache_prune_timeout_id = 0;
+}
+
+static GClueLocationSourceStartResult
 gclue_wifi_start (GClueLocationSource *source)
 {
         GClueLocationSourceClass *base_class;
+        GClueLocationSourceStartResult base_result;
 
-        g_return_val_if_fail (GCLUE_IS_WIFI (source), FALSE);
+        g_return_val_if_fail (GCLUE_IS_WIFI (source),
+                              GCLUE_LOCATION_SOURCE_START_RESULT_FAILED);
 
         base_class = GCLUE_LOCATION_SOURCE_CLASS (gclue_wifi_parent_class);
-        if (!base_class->start (source))
-                return FALSE;
+        base_result = base_class->start (source);
+        if (base_result != GCLUE_LOCATION_SOURCE_START_RESULT_OK)
+                return base_result;
 
+        connect_cache_prune_timeout (GCLUE_WIFI (source));
         connect_bss_signals (GCLUE_WIFI (source));
-        return TRUE;
+
+        return base_result;
 }
 
-static gboolean
+static GClueLocationSourceStopResult
 gclue_wifi_stop (GClueLocationSource *source)
 {
         GClueLocationSourceClass *base_class;
+        GClueLocationSourceStopResult base_result;
 
         g_return_val_if_fail (GCLUE_IS_WIFI (source), FALSE);
 
         base_class = GCLUE_LOCATION_SOURCE_CLASS (gclue_wifi_parent_class);
-        if (!base_class->stop (source))
-                return FALSE;
+        base_result = base_class->stop (source);
+        if (base_result == GCLUE_LOCATION_SOURCE_STOP_RESULT_STILL_USED)
+                return base_result;
 
         disconnect_bss_signals (GCLUE_WIFI (source));
-        return TRUE;
+        disconnect_cache_prune_timeout (GCLUE_WIFI (source));
+
+        return base_result;
 }
 
 static GClueAccuracyLevel
@@ -712,7 +894,7 @@ on_interface_removed (WPASupplicant *supplicant,
 static void
 gclue_wifi_init (GClueWifi *wifi)
 {
-        wifi->priv = G_TYPE_INSTANCE_GET_PRIVATE ((wifi), GCLUE_TYPE_WIFI, GClueWifiPrivate);
+        wifi->priv = gclue_wifi_get_instance_private (wifi);
 
         wifi->priv->bss_proxies = g_hash_table_new_full (g_str_hash,
                                                          g_str_equal,
@@ -722,6 +904,10 @@ gclue_wifi_init (GClueWifi *wifi)
                                                                  g_str_equal,
                                                                  g_free,
                                                                  g_object_unref);
+        wifi->priv->location_cache = g_hash_table_new_full (variant_hash,
+                                                            g_variant_equal,
+                                                            (GDestroyNotify) g_variant_unref,
+                                                            g_object_unref);
 }
 
 static void
@@ -799,6 +985,7 @@ gclue_wifi_get_singleton (GClueAccuracyLevel level)
         static GClueWifi *wifi[] = { NULL, NULL };
         guint i;
         gboolean scramble_location = FALSE;
+        gboolean compute_movement = FALSE;
 
         g_return_val_if_fail (level >= GCLUE_ACCURACY_LEVEL_CITY, NULL);
         if (level == GCLUE_ACCURACY_LEVEL_NEIGHBORHOOD)
@@ -812,12 +999,14 @@ gclue_wifi_get_singleton (GClueAccuracyLevel level)
                         scramble_location = TRUE;
         } else {
                 i = 1;
+                compute_movement = TRUE;
         }
 
         if (wifi[i] == NULL) {
                 wifi[i] = g_object_new (GCLUE_TYPE_WIFI,
                                         "accuracy-level", level,
                                         "scramble-location", scramble_location,
+                                        "compute-movement", compute_movement,
                                         NULL);
                 g_object_weak_ref (G_OBJECT (wifi[i]),
                                    on_wifi_destroyed,
@@ -837,19 +1026,10 @@ gclue_wifi_get_accuracy_level (GClueWifi *wifi)
         return wifi->priv->accuracy_level;
 }
 
-/* Can return NULL without setting @error, signifying an empty BSS list. */
+/* Can return NULL, signifying an empty BSS list. */
 static GList *
-get_bss_list (GClueWifi *wifi,
-              GError   **error)
+get_bss_list (GClueWifi *wifi)
 {
-        if (wifi->priv->interface == NULL) {
-                g_set_error_literal (error,
-                                     G_IO_ERROR,
-                                     G_IO_ERROR_FAILED,
-                                     "No WiFi devices available");
-                return NULL;
-        }
-
         return g_hash_table_get_values (wifi->priv->bss_proxies);
 }
 
@@ -857,16 +1037,16 @@ static SoupMessage *
 gclue_wifi_create_query (GClueWebSource *source,
                          GError        **error)
 {
-        GList *bss_list; /* As in Access Points */
+        GClueWifi *wifi = GCLUE_WIFI (source);
+        GList *bss_list = NULL; /* As in Access Points */
         SoupMessage *msg;
-        g_autoptr(GError) local_error = NULL;
 
-        bss_list = get_bss_list (GCLUE_WIFI (source), &local_error);
-        if (local_error != NULL) {
-                g_propagate_error (error, g_steal_pointer (&local_error));
-                return NULL;
+        if (wifi->priv->interface == NULL) {
+                goto create_query;
         }
 
+        bss_list = get_bss_list (wifi);
+
         /* Empty list? */
         if (bss_list == NULL) {
                 g_set_error_literal (error,
@@ -876,6 +1056,7 @@ gclue_wifi_create_query (GClueWebSource *source,
                 return NULL;
         }
 
+create_query:
         msg = gclue_mozilla_create_query (bss_list, NULL, error);
         g_list_free (bss_list);
         return msg;
@@ -894,16 +1075,20 @@ gclue_wifi_create_submit_query (GClueWebSource  *source,
                                 GClueLocation   *location,
                                 GError         **error)
 {
+        GClueWifi *wifi = GCLUE_WIFI (source);
         GList *bss_list; /* As in Access Points */
         SoupMessage * msg;
-        g_autoptr(GError) local_error = NULL;
 
-        bss_list = get_bss_list (GCLUE_WIFI (source), &local_error);
-        if (local_error != NULL) {
-                g_propagate_error (error, g_steal_pointer (&local_error));
+        if (wifi->priv->interface == NULL) {
+                g_set_error_literal (error,
+                                     G_IO_ERROR,
+                                     G_IO_ERROR_FAILED,
+                                     "No WiFi devices available");
                 return NULL;
         }
 
+        bss_list = get_bss_list (wifi);
+
         /* Empty list? */
         if (bss_list == NULL) {
                 g_set_error_literal (error,
@@ -920,3 +1105,198 @@ gclue_wifi_create_submit_query (GClueWebSource  *source,
         g_list_free (bss_list);
         return msg;
 }
+
+static void refresh_cb (GObject      *source_object,
+                        GAsyncResult *result,
+                        gpointer      user_data);
+
+static gint
+bss_compare (gconstpointer a,
+             gconstpointer b)
+{
+        WPABSS **_bss_a = (WPABSS **) a;
+        WPABSS **_bss_b = (WPABSS **) b;
+        WPABSS *bss_a = WPA_BSS (*_bss_a);
+        WPABSS *bss_b = WPA_BSS (*_bss_b);
+        GVariant *bssid_a = wpa_bss_get_bssid (bss_a);
+        GVariant *bssid_b = wpa_bss_get_bssid (bss_b);
+        g_autoptr(GBytes) bssid_bytes_a = NULL;
+        g_autoptr(GBytes) bssid_bytes_b = NULL;
+
+        /* Can’t use g_variant_compare() as it isn’t defined over `ay` types */
+        if (bssid_a == NULL && bssid_b == NULL)
+                return 0;
+        else if (bssid_a == NULL)
+                return -1;
+        else if (bssid_b == NULL)
+                return 1;
+
+        bssid_bytes_a = g_variant_get_data_as_bytes (bssid_a);
+        bssid_bytes_b = g_variant_get_data_as_bytes (bssid_b);
+
+        return g_bytes_compare (bssid_bytes_a, bssid_bytes_b);
+}
+
+static guint
+variant_hash (gconstpointer key)
+{
+        GVariant *variant = (GVariant *) key;
+        g_autoptr(GBytes) bytes = g_variant_get_data_as_bytes (variant);
+        return g_bytes_hash (bytes);
+}
+
+static GVariant *
+get_location_cache_key (GClueWifi *wifi)
+{
+        GHashTableIter iter;
+        gpointer value;
+        g_autoptr(GPtrArray) bss_array = g_ptr_array_new_with_free_func (NULL);  /* (element-type WPABSS) */
+        guint i;
+        GVariantBuilder builder;
+
+        /* The Mozilla service puts BSSID and signal strength for each BSS into
+         * its query. The signal strength can typically vary by ±5 for a
+         * stationary laptop, so quantise by that. Pack the whole lot into a
+         * #GVariant for simplicity, sorted by MAC address. The sorting has to
+         * happen in an array beforehand, as variants are immutable. */
+        g_hash_table_iter_init (&iter, wifi->priv->bss_proxies);
+
+        while (g_hash_table_iter_next (&iter, NULL, &value)) {
+                WPABSS *bss = WPA_BSS (value);
+                if (bss != NULL)
+                        g_ptr_array_add (bss_array, bss);
+        }
+
+        g_ptr_array_sort (bss_array, bss_compare);
+
+        /* Serialise to a variant. */
+        g_variant_builder_init (&builder, G_VARIANT_TYPE ("a(ayn)"));
+        for (i = 0; i < bss_array->len; i++) {
+                WPABSS *bss = WPA_BSS (bss_array->pdata[i]);
+                GVariant *bssid;
+
+                g_variant_builder_open (&builder, G_VARIANT_TYPE ("(ayn)"));
+
+                bssid = wpa_bss_get_bssid (bss);
+                if (bssid == NULL)
+                        continue;
+
+                g_variant_builder_add_value (&builder, bssid);
+                g_variant_builder_add (&builder, "n", wpa_bss_get_signal (bss) / 10);
+
+                g_variant_builder_close (&builder);
+        }
+
+        return g_variant_builder_end (&builder);
+}
+
+static GClueLocation *
+duplicate_location_new_timestamp (GClueLocation *location)
+{
+        return g_object_new (GCLUE_TYPE_LOCATION,
+                             "latitude", gclue_location_get_latitude (location),
+                             "longitude", gclue_location_get_longitude (location),
+                             "accuracy", gclue_location_get_accuracy (location),
+                             "altitude", gclue_location_get_altitude (location),
+                             "timestamp", 0,
+                             "speed", gclue_location_get_speed (location),
+                             "heading", gclue_location_get_heading (location),
+                             NULL);
+}
+
+static void
+gclue_wifi_refresh_async (GClueWebSource      *source,
+                          GCancellable        *cancellable,
+                          GAsyncReadyCallback  callback,
+                          gpointer             user_data)
+{
+        GClueWifi *wifi = GCLUE_WIFI (source);
+        g_autoptr(GTask) task = g_task_new (source, cancellable, callback, user_data);
+        g_autoptr(GVariant) cache_key = get_location_cache_key (wifi);
+        g_autofree gchar *cache_key_str = g_variant_print (cache_key, FALSE);
+        GClueLocation *cached_location = g_hash_table_lookup (wifi->priv->location_cache, cache_key);
+
+        g_task_set_source_tag (task, gclue_wifi_refresh_async);
+        g_task_set_task_data (task, g_steal_pointer (&cache_key), (GDestroyNotify) g_variant_unref);
+
+        if (gclue_location_source_get_active (GCLUE_LOCATION_SOURCE (source))) {
+                /* Try the cache. */
+                if (cached_location != NULL) {
+                        g_autoptr(GClueLocation) new_location = NULL;
+
+                        g_debug ("Cache hit for key %s: got location %p (%s)",
+                                 cache_key_str, cached_location,
+                                 gclue_location_get_description (cached_location));
+                        wifi->priv->cache_hits++;
+
+                        /* Duplicate the location so its timestamp is updated. */
+                        new_location = duplicate_location_new_timestamp (cached_location);
+                        gclue_location_source_set_location (GCLUE_LOCATION_SOURCE (source), new_location);
+
+                        g_task_return_pointer (task, g_steal_pointer (&new_location), g_object_unref);
+                        return;
+                }
+
+                g_debug ("Cache miss for key %s; querying web service", cache_key_str);
+                wifi->priv->cache_misses++;
+        }
+
+        /* Fall back to querying the web service. */
+        GCLUE_WEB_SOURCE_CLASS (gclue_wifi_parent_class)->refresh_async (source, cancellable, refresh_cb, g_steal_pointer (&task));
+}
+
+static void
+refresh_cb (GObject      *source_object,
+            GAsyncResult *result,
+            gpointer      user_data)
+{
+        GClueWebSource *source = GCLUE_WEB_SOURCE (source_object);
+        GClueWifi *wifi = GCLUE_WIFI (source);
+        g_autoptr(GTask) task = g_steal_pointer (&user_data);
+        g_autoptr(GClueLocation) location = NULL;
+        g_autoptr(GError) local_error = NULL;
+        GVariant *cache_key;
+        g_autofree gchar *cache_key_str = NULL;
+        double cache_hit_ratio;
+
+        /* Finish querying the web service. */
+        location = GCLUE_WEB_SOURCE_CLASS (gclue_wifi_parent_class)->refresh_finish (source, result, &local_error);
+
+        if (local_error != NULL) {
+                g_task_return_error (task, g_steal_pointer (&local_error));
+                return;
+        }
+
+        /* Cache the result. */
+        cache_key = g_task_get_task_data (task);
+        cache_key_str = g_variant_print (cache_key, FALSE);
+        g_hash_table_replace (wifi->priv->location_cache, g_variant_ref (cache_key), g_object_ref (location));
+
+        if (wifi->priv->cache_hits || wifi->priv->cache_misses) {
+                double cache_attempts;
+
+                cache_attempts = wifi->priv->cache_hits;
+                cache_attempts += wifi->priv->cache_misses;
+                cache_hit_ratio = wifi->priv->cache_hits * 100.0 / cache_attempts;
+        } else {
+                cache_hit_ratio = 0;
+        }
+
+        g_debug ("Adding %s / %s to cache (new size: %u; hit ratio %.2f%%)",
+                 cache_key_str,
+                 gclue_location_get_description (location),
+                 g_hash_table_size (wifi->priv->location_cache),
+                 cache_hit_ratio);
+
+        g_task_return_pointer (task, g_steal_pointer (&location), g_object_unref);
+}
+
+static GClueLocation *
+gclue_wifi_refresh_finish (GClueWebSource  *source,
+                           GAsyncResult    *result,
+                           GError         **error)
+{
+        GTask *task = G_TASK (result);
+
+        return g_task_propagate_pointer (task, error);
+}
